# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Annotate opcode and encoding fmt using CoreDSL2 attributes."""

import sys
import argparse
import logging
import pathlib

import pandas as pd

import seal5.model as seal5_model
from seal5.model_utils import load_model, dump_model
from seal5.riscv_utils import detect_opcode, detect_format, detect_funct3_funct7

from m2isar.metamodel import behav

logger = logging.getLogger("annotate_opcodes")


def annotate_opcodes(instr_def: seal5_model.Seal5Instruction, enc_mask: int, enc_match: int):
    enc_size = instr_def.size
    opcode_name, opcode_val = detect_opcode(instr_def)
    # print("opcode_name", opcode_name)
    # print("opcode_val", opcode_val)
    funct3_val, funct7_val = detect_funct3_funct7(instr_def)
    # print("funct3_val", funct3_val)
    # print("funct7_val", funct7_val)
    enc_format, enc_pattern = detect_format(instr_def)
    # print("enc_format", enc_format)
    # print("enc_pattern", enc_pattern)
    if opcode_val is not None:
        # TODO: IntLiteral?
        instr_def.attributes[seal5_model.Seal5InstrAttribute.OPCODE] = behav.IntLiteral(opcode_val, 7)
    if opcode_name is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.OPCODE_NAME] = behav.StringLiteral(opcode_name)
    if funct3_val is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.FUNCT3] = behav.IntLiteral(funct3_val, 3)
    if funct7_val is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.FUNCT7] = behav.IntLiteral(funct7_val, 7)
    if enc_format is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.ENC_FORMAT] = behav.StringLiteral(enc_format)
    if enc_pattern is not None:
        # TODO: StringRef?
        instr_def.attributes[seal5_model.Seal5InstrAttribute.ENC_PATTERN] = behav.StringLiteral(enc_pattern)
    if enc_mask is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.ENC_MASK] = behav.IntLiteral(enc_mask, enc_size)
    if enc_match is not None:
        instr_def.attributes[seal5_model.Seal5InstrAttribute.ENC_MATCH] = behav.IntLiteral(enc_match, enc_size)


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }

    for set_name, set_def in model_obj.sets.items():
        metrics["n_sets"] += 1
        logger.info("processing set %s", set_name)
        for key, instr_def in set_def.instructions.items():
            metrics["n_instructions"] += 1
            logger.info("processing instruction %s", instr_def.name)
            enc_match, enc_mask = key
            # print("mask", hex(enc_mask))
            # print("match_", hex(enc_match))
            try:
                annotate_opcodes(instr_def, enc_mask, enc_match)
                metrics["n_success"] += 1
                metrics["success_instructions"].append(instr_def.name)
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_instructions"].append(instr_def.name)

    dump_model(model_obj, out_path, compat=args.compat)
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Classes to recursively import files of a CoreDSL model."""

import pathlib

from .parser_gen import CoreDSL2Listener, CoreDSL2Parser, CoreDSL2Visitor
from .utils import make_parser
from seal5.logging import Logger


class Importer(CoreDSL2Listener):
    """ANTLR listener based importer. Bad on performance, as it traverses
    the entire parse tree when it only has to look for import statements.
    """

    def __init__(self, search_path) -> None:
        super().__init__()
        self.imported = set()
        self.new_children = []
        self.new_defs = []
        self.got_new = True
        self.search_path = search_path

    def enterImport_file(self, ctx: CoreDSL2Parser.Import_fileContext):
        """The actual import functionality. Extracts the filename to import,
        constructs a new parser and parses the next file.
        """

        filename = ctx.RULE_STRING().getText().replace('"', "")
        if filename not in self.imported:
            print(f"importing file {filename}")
            self.got_new = True
            self.imported.add(filename)

            parser = make_parser(self.search_path / filename)

            tree = parser.description_content()

            self.new_children.extend(tree.children)
            self.new_defs.extend(tree.definitions)
        pass


def recursive_import(tree, search_path):
    """Helper method to recursively process all import statements of a given
    parse tree. The search path should be set to the directory of the root document.
    """

    path_extender = ImportPathExtender(search_path)
    path_extender.visit(tree)

    importer = VisitImporter(search_path)

    while importer.got_new:
        importer.new_imports.clear()
        importer.new_defs.clear()
        importer.new_children.clear()
        importer.got_new = False

        importer.visit(tree)

        tree.imports = importer.new_imports + tree.imports
        tree.definitions = importer.new_defs + tree.definitions
        tree.children = importer.new_children + [
            x for x in tree.children if not isinstance(x, CoreDSL2Parser.Import_fileContext)
        ]


class VisitImporter(CoreDSL2Visitor):
    """Importer class based on an ANTLR Visitor. Only traverses the model tree
    to the import statements and stops traversion after that.
    """

    def __init__(self, search_path) -> None:
        super().__init__()
        self.imported = set()
        self.new_children = []
        self.new_imports = []
        self.new_defs = []
        self.got_new = True
        self.search_path = search_path
        self.logger = Logger("frontends.visit_importer")

    def visitDescription_content(self, ctx: CoreDSL2Parser.Description_contentContext):
        for i in ctx.imports:
            self.visit(i)

    def visitImport_file(self, ctx: CoreDSL2Parser.Import_fileContext):
        """The actual import functionality. Extracts the filename to import,
        constructs a new parser and parses the next file.
        """

        import_name = ctx.uri.text.replace('"', "")
        filename = str(pathlib.Path(import_name).resolve())

        # only import each file once
        if filename not in self.imported:
            self.logger.info("importing file %s", filename)

            # keep track that we imported something
            self.got_new = True
            self.imported.add(filename)

            # extract file path and search path
            file_path = pathlib.Path(filename)
            file_dir = file_path.parent

            parser = make_parser(file_path)

            # run ImportPathExtender on the new tree
            tree = parser.description_content()
            path_extender = ImportPathExtender(file_dir)
            path_extender.visit(tree)

            # keep track of the new children
            self.new_children.extend(tree.children)
            self.new_imports.extend(tree.imports)
            self.new_defs.extend(tree.definitions)


class ImportPathExtender(CoreDSL2Visitor):
    """ANTLR visitor to resolve relative import paths. Replaces all import URIs
    with their equivalent absolute path, relative to search_path.
    """

    def __init__(self, search_path: pathlib.Path) -> None:
        super().__init__()
        self.search_path = search_path

    def visitDescription_content(self, ctx: CoreDSL2Parser.Description_contentContext):
        for i in ctx.imports:
            self.visit(i)

    def visitImport_file(self, ctx: CoreDSL2Parser.Import_fileContext):
        filename = self.search_path / ctx.uri.text.replace('"', "")
        ctx.uri.text = str(filename)



# Generated from CoreDSL2.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .CoreDSL2Parser import CoreDSL2Parser
else:
    from CoreDSL2Parser import CoreDSL2Parser

# This class defines a complete listener for a parse tree produced by CoreDSL2Parser.
class CoreDSL2Listener(ParseTreeListener):

    # Enter a parse tree produced by CoreDSL2Parser#description_content.
    def enterDescription_content(self, ctx:CoreDSL2Parser.Description_contentContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#description_content.
    def exitDescription_content(self, ctx:CoreDSL2Parser.Description_contentContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#import_file.
    def enterImport_file(self, ctx:CoreDSL2Parser.Import_fileContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#import_file.
    def exitImport_file(self, ctx:CoreDSL2Parser.Import_fileContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#instruction_set.
    def enterInstruction_set(self, ctx:CoreDSL2Parser.Instruction_setContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#instruction_set.
    def exitInstruction_set(self, ctx:CoreDSL2Parser.Instruction_setContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#core_def.
    def enterCore_def(self, ctx:CoreDSL2Parser.Core_defContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#core_def.
    def exitCore_def(self, ctx:CoreDSL2Parser.Core_defContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#section_arch_state.
    def enterSection_arch_state(self, ctx:CoreDSL2Parser.Section_arch_stateContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#section_arch_state.
    def exitSection_arch_state(self, ctx:CoreDSL2Parser.Section_arch_stateContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#section_functions.
    def enterSection_functions(self, ctx:CoreDSL2Parser.Section_functionsContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#section_functions.
    def exitSection_functions(self, ctx:CoreDSL2Parser.Section_functionsContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#section_instructions.
    def enterSection_instructions(self, ctx:CoreDSL2Parser.Section_instructionsContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#section_instructions.
    def exitSection_instructions(self, ctx:CoreDSL2Parser.Section_instructionsContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#section_always.
    def enterSection_always(self, ctx:CoreDSL2Parser.Section_alwaysContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#section_always.
    def exitSection_always(self, ctx:CoreDSL2Parser.Section_alwaysContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#always_block.
    def enterAlways_block(self, ctx:CoreDSL2Parser.Always_blockContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#always_block.
    def exitAlways_block(self, ctx:CoreDSL2Parser.Always_blockContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#instruction.
    def enterInstruction(self, ctx:CoreDSL2Parser.InstructionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#instruction.
    def exitInstruction(self, ctx:CoreDSL2Parser.InstructionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#rule_encoding.
    def enterRule_encoding(self, ctx:CoreDSL2Parser.Rule_encodingContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#rule_encoding.
    def exitRule_encoding(self, ctx:CoreDSL2Parser.Rule_encodingContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#bit_value.
    def enterBit_value(self, ctx:CoreDSL2Parser.Bit_valueContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#bit_value.
    def exitBit_value(self, ctx:CoreDSL2Parser.Bit_valueContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#bit_field.
    def enterBit_field(self, ctx:CoreDSL2Parser.Bit_fieldContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#bit_field.
    def exitBit_field(self, ctx:CoreDSL2Parser.Bit_fieldContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#function_definition.
    def enterFunction_definition(self, ctx:CoreDSL2Parser.Function_definitionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#function_definition.
    def exitFunction_definition(self, ctx:CoreDSL2Parser.Function_definitionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#parameter_list.
    def enterParameter_list(self, ctx:CoreDSL2Parser.Parameter_listContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#parameter_list.
    def exitParameter_list(self, ctx:CoreDSL2Parser.Parameter_listContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#parameter_declaration.
    def enterParameter_declaration(self, ctx:CoreDSL2Parser.Parameter_declarationContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#parameter_declaration.
    def exitParameter_declaration(self, ctx:CoreDSL2Parser.Parameter_declarationContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#block_statement.
    def enterBlock_statement(self, ctx:CoreDSL2Parser.Block_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#block_statement.
    def exitBlock_statement(self, ctx:CoreDSL2Parser.Block_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#procedure_call.
    def enterProcedure_call(self, ctx:CoreDSL2Parser.Procedure_callContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#procedure_call.
    def exitProcedure_call(self, ctx:CoreDSL2Parser.Procedure_callContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#if_statement.
    def enterIf_statement(self, ctx:CoreDSL2Parser.If_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#if_statement.
    def exitIf_statement(self, ctx:CoreDSL2Parser.If_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#for_statement.
    def enterFor_statement(self, ctx:CoreDSL2Parser.For_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#for_statement.
    def exitFor_statement(self, ctx:CoreDSL2Parser.For_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#while_statement.
    def enterWhile_statement(self, ctx:CoreDSL2Parser.While_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#while_statement.
    def exitWhile_statement(self, ctx:CoreDSL2Parser.While_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#do_statement.
    def enterDo_statement(self, ctx:CoreDSL2Parser.Do_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#do_statement.
    def exitDo_statement(self, ctx:CoreDSL2Parser.Do_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#switch_statement.
    def enterSwitch_statement(self, ctx:CoreDSL2Parser.Switch_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#switch_statement.
    def exitSwitch_statement(self, ctx:CoreDSL2Parser.Switch_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#return_statement.
    def enterReturn_statement(self, ctx:CoreDSL2Parser.Return_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#return_statement.
    def exitReturn_statement(self, ctx:CoreDSL2Parser.Return_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#break_statement.
    def enterBreak_statement(self, ctx:CoreDSL2Parser.Break_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#break_statement.
    def exitBreak_statement(self, ctx:CoreDSL2Parser.Break_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#continue_statement.
    def enterContinue_statement(self, ctx:CoreDSL2Parser.Continue_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#continue_statement.
    def exitContinue_statement(self, ctx:CoreDSL2Parser.Continue_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#spawn_statement.
    def enterSpawn_statement(self, ctx:CoreDSL2Parser.Spawn_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#spawn_statement.
    def exitSpawn_statement(self, ctx:CoreDSL2Parser.Spawn_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#expression_statement.
    def enterExpression_statement(self, ctx:CoreDSL2Parser.Expression_statementContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#expression_statement.
    def exitExpression_statement(self, ctx:CoreDSL2Parser.Expression_statementContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#switch_block_statement_group.
    def enterSwitch_block_statement_group(self, ctx:CoreDSL2Parser.Switch_block_statement_groupContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#switch_block_statement_group.
    def exitSwitch_block_statement_group(self, ctx:CoreDSL2Parser.Switch_block_statement_groupContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#switch_label.
    def enterSwitch_label(self, ctx:CoreDSL2Parser.Switch_labelContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#switch_label.
    def exitSwitch_label(self, ctx:CoreDSL2Parser.Switch_labelContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#block.
    def enterBlock(self, ctx:CoreDSL2Parser.BlockContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#block.
    def exitBlock(self, ctx:CoreDSL2Parser.BlockContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#block_item.
    def enterBlock_item(self, ctx:CoreDSL2Parser.Block_itemContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#block_item.
    def exitBlock_item(self, ctx:CoreDSL2Parser.Block_itemContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#for_condition.
    def enterFor_condition(self, ctx:CoreDSL2Parser.For_conditionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#for_condition.
    def exitFor_condition(self, ctx:CoreDSL2Parser.For_conditionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#declaration.
    def enterDeclaration(self, ctx:CoreDSL2Parser.DeclarationContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#declaration.
    def exitDeclaration(self, ctx:CoreDSL2Parser.DeclarationContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#type_specifier.
    def enterType_specifier(self, ctx:CoreDSL2Parser.Type_specifierContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#type_specifier.
    def exitType_specifier(self, ctx:CoreDSL2Parser.Type_specifierContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#integer_type.
    def enterInteger_type(self, ctx:CoreDSL2Parser.Integer_typeContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#integer_type.
    def exitInteger_type(self, ctx:CoreDSL2Parser.Integer_typeContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#float_type.
    def enterFloat_type(self, ctx:CoreDSL2Parser.Float_typeContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#float_type.
    def exitFloat_type(self, ctx:CoreDSL2Parser.Float_typeContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#bool_type.
    def enterBool_type(self, ctx:CoreDSL2Parser.Bool_typeContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#bool_type.
    def exitBool_type(self, ctx:CoreDSL2Parser.Bool_typeContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#void_type.
    def enterVoid_type(self, ctx:CoreDSL2Parser.Void_typeContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#void_type.
    def exitVoid_type(self, ctx:CoreDSL2Parser.Void_typeContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#composite_declaration.
    def enterComposite_declaration(self, ctx:CoreDSL2Parser.Composite_declarationContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#composite_declaration.
    def exitComposite_declaration(self, ctx:CoreDSL2Parser.Composite_declarationContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#composite_reference.
    def enterComposite_reference(self, ctx:CoreDSL2Parser.Composite_referenceContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#composite_reference.
    def exitComposite_reference(self, ctx:CoreDSL2Parser.Composite_referenceContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#enum_declaration.
    def enterEnum_declaration(self, ctx:CoreDSL2Parser.Enum_declarationContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#enum_declaration.
    def exitEnum_declaration(self, ctx:CoreDSL2Parser.Enum_declarationContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#enum_reference.
    def enterEnum_reference(self, ctx:CoreDSL2Parser.Enum_referenceContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#enum_reference.
    def exitEnum_reference(self, ctx:CoreDSL2Parser.Enum_referenceContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#integer_signedness.
    def enterInteger_signedness(self, ctx:CoreDSL2Parser.Integer_signednessContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#integer_signedness.
    def exitInteger_signedness(self, ctx:CoreDSL2Parser.Integer_signednessContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#integer_shorthand.
    def enterInteger_shorthand(self, ctx:CoreDSL2Parser.Integer_shorthandContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#integer_shorthand.
    def exitInteger_shorthand(self, ctx:CoreDSL2Parser.Integer_shorthandContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#float_shorthand.
    def enterFloat_shorthand(self, ctx:CoreDSL2Parser.Float_shorthandContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#float_shorthand.
    def exitFloat_shorthand(self, ctx:CoreDSL2Parser.Float_shorthandContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#attribute.
    def enterAttribute(self, ctx:CoreDSL2Parser.AttributeContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#attribute.
    def exitAttribute(self, ctx:CoreDSL2Parser.AttributeContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#bit_size_specifier.
    def enterBit_size_specifier(self, ctx:CoreDSL2Parser.Bit_size_specifierContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#bit_size_specifier.
    def exitBit_size_specifier(self, ctx:CoreDSL2Parser.Bit_size_specifierContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#enumerator_list.
    def enterEnumerator_list(self, ctx:CoreDSL2Parser.Enumerator_listContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#enumerator_list.
    def exitEnumerator_list(self, ctx:CoreDSL2Parser.Enumerator_listContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#enumerator.
    def enterEnumerator(self, ctx:CoreDSL2Parser.EnumeratorContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#enumerator.
    def exitEnumerator(self, ctx:CoreDSL2Parser.EnumeratorContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#struct_declaration.
    def enterStruct_declaration(self, ctx:CoreDSL2Parser.Struct_declarationContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#struct_declaration.
    def exitStruct_declaration(self, ctx:CoreDSL2Parser.Struct_declarationContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#struct_declaration_specifier.
    def enterStruct_declaration_specifier(self, ctx:CoreDSL2Parser.Struct_declaration_specifierContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#struct_declaration_specifier.
    def exitStruct_declaration_specifier(self, ctx:CoreDSL2Parser.Struct_declaration_specifierContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#declarator.
    def enterDeclarator(self, ctx:CoreDSL2Parser.DeclaratorContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#declarator.
    def exitDeclarator(self, ctx:CoreDSL2Parser.DeclaratorContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#initializer.
    def enterInitializer(self, ctx:CoreDSL2Parser.InitializerContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#initializer.
    def exitInitializer(self, ctx:CoreDSL2Parser.InitializerContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#initializerList.
    def enterInitializerList(self, ctx:CoreDSL2Parser.InitializerListContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#initializerList.
    def exitInitializerList(self, ctx:CoreDSL2Parser.InitializerListContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#designated_initializer.
    def enterDesignated_initializer(self, ctx:CoreDSL2Parser.Designated_initializerContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#designated_initializer.
    def exitDesignated_initializer(self, ctx:CoreDSL2Parser.Designated_initializerContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#designator.
    def enterDesignator(self, ctx:CoreDSL2Parser.DesignatorContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#designator.
    def exitDesignator(self, ctx:CoreDSL2Parser.DesignatorContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#cast_expression.
    def enterCast_expression(self, ctx:CoreDSL2Parser.Cast_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#cast_expression.
    def exitCast_expression(self, ctx:CoreDSL2Parser.Cast_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#binary_expression.
    def enterBinary_expression(self, ctx:CoreDSL2Parser.Binary_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#binary_expression.
    def exitBinary_expression(self, ctx:CoreDSL2Parser.Binary_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#preinc_expression.
    def enterPreinc_expression(self, ctx:CoreDSL2Parser.Preinc_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#preinc_expression.
    def exitPreinc_expression(self, ctx:CoreDSL2Parser.Preinc_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#conditional_expression.
    def enterConditional_expression(self, ctx:CoreDSL2Parser.Conditional_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#conditional_expression.
    def exitConditional_expression(self, ctx:CoreDSL2Parser.Conditional_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#deref_expression.
    def enterDeref_expression(self, ctx:CoreDSL2Parser.Deref_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#deref_expression.
    def exitDeref_expression(self, ctx:CoreDSL2Parser.Deref_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#prefix_expression.
    def enterPrefix_expression(self, ctx:CoreDSL2Parser.Prefix_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#prefix_expression.
    def exitPrefix_expression(self, ctx:CoreDSL2Parser.Prefix_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#postinc_expression.
    def enterPostinc_expression(self, ctx:CoreDSL2Parser.Postinc_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#postinc_expression.
    def exitPostinc_expression(self, ctx:CoreDSL2Parser.Postinc_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#concat_expression.
    def enterConcat_expression(self, ctx:CoreDSL2Parser.Concat_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#concat_expression.
    def exitConcat_expression(self, ctx:CoreDSL2Parser.Concat_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#assignment_expression.
    def enterAssignment_expression(self, ctx:CoreDSL2Parser.Assignment_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#assignment_expression.
    def exitAssignment_expression(self, ctx:CoreDSL2Parser.Assignment_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#method_call.
    def enterMethod_call(self, ctx:CoreDSL2Parser.Method_callContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#method_call.
    def exitMethod_call(self, ctx:CoreDSL2Parser.Method_callContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#primary_expression.
    def enterPrimary_expression(self, ctx:CoreDSL2Parser.Primary_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#primary_expression.
    def exitPrimary_expression(self, ctx:CoreDSL2Parser.Primary_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#slice_expression.
    def enterSlice_expression(self, ctx:CoreDSL2Parser.Slice_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#slice_expression.
    def exitSlice_expression(self, ctx:CoreDSL2Parser.Slice_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#reference_expression.
    def enterReference_expression(self, ctx:CoreDSL2Parser.Reference_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#reference_expression.
    def exitReference_expression(self, ctx:CoreDSL2Parser.Reference_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#constant_expression.
    def enterConstant_expression(self, ctx:CoreDSL2Parser.Constant_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#constant_expression.
    def exitConstant_expression(self, ctx:CoreDSL2Parser.Constant_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#literal_expression.
    def enterLiteral_expression(self, ctx:CoreDSL2Parser.Literal_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#literal_expression.
    def exitLiteral_expression(self, ctx:CoreDSL2Parser.Literal_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#parens_expression.
    def enterParens_expression(self, ctx:CoreDSL2Parser.Parens_expressionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#parens_expression.
    def exitParens_expression(self, ctx:CoreDSL2Parser.Parens_expressionContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#string_literal.
    def enterString_literal(self, ctx:CoreDSL2Parser.String_literalContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#string_literal.
    def exitString_literal(self, ctx:CoreDSL2Parser.String_literalContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#constant.
    def enterConstant(self, ctx:CoreDSL2Parser.ConstantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#constant.
    def exitConstant(self, ctx:CoreDSL2Parser.ConstantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#integer_constant.
    def enterInteger_constant(self, ctx:CoreDSL2Parser.Integer_constantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#integer_constant.
    def exitInteger_constant(self, ctx:CoreDSL2Parser.Integer_constantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#floating_constant.
    def enterFloating_constant(self, ctx:CoreDSL2Parser.Floating_constantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#floating_constant.
    def exitFloating_constant(self, ctx:CoreDSL2Parser.Floating_constantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#bool_constant.
    def enterBool_constant(self, ctx:CoreDSL2Parser.Bool_constantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#bool_constant.
    def exitBool_constant(self, ctx:CoreDSL2Parser.Bool_constantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#character_constant.
    def enterCharacter_constant(self, ctx:CoreDSL2Parser.Character_constantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#character_constant.
    def exitCharacter_constant(self, ctx:CoreDSL2Parser.Character_constantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#string_constant.
    def enterString_constant(self, ctx:CoreDSL2Parser.String_constantContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#string_constant.
    def exitString_constant(self, ctx:CoreDSL2Parser.String_constantContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#double_left_bracket.
    def enterDouble_left_bracket(self, ctx:CoreDSL2Parser.Double_left_bracketContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#double_left_bracket.
    def exitDouble_left_bracket(self, ctx:CoreDSL2Parser.Double_left_bracketContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#double_right_bracket.
    def enterDouble_right_bracket(self, ctx:CoreDSL2Parser.Double_right_bracketContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#double_right_bracket.
    def exitDouble_right_bracket(self, ctx:CoreDSL2Parser.Double_right_bracketContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#data_types.
    def enterData_types(self, ctx:CoreDSL2Parser.Data_typesContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#data_types.
    def exitData_types(self, ctx:CoreDSL2Parser.Data_typesContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#type_qualifier.
    def enterType_qualifier(self, ctx:CoreDSL2Parser.Type_qualifierContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#type_qualifier.
    def exitType_qualifier(self, ctx:CoreDSL2Parser.Type_qualifierContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#storage_class_specifier.
    def enterStorage_class_specifier(self, ctx:CoreDSL2Parser.Storage_class_specifierContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#storage_class_specifier.
    def exitStorage_class_specifier(self, ctx:CoreDSL2Parser.Storage_class_specifierContext):
        pass


    # Enter a parse tree produced by CoreDSL2Parser#struct_or_union.
    def enterStruct_or_union(self, ctx:CoreDSL2Parser.Struct_or_unionContext):
        pass

    # Exit a parse tree produced by CoreDSL2Parser#struct_or_union.
    def exitStruct_or_union(self, ctx:CoreDSL2Parser.Struct_or_unionContext):
        pass



del CoreDSL2Parser


# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Status report writer for Seal5."""

import argparse
import logging
import pathlib

import pandas as pd

from seal5.settings import Seal5Settings
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.status_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", nargs="+", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md"], default="auto")
    parser.add_argument("--yaml", type=str, default=None)
    parser.add_argument("--compact", action="store_true")
    parser.add_argument("--markdown-icons", action="store_true")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # if args.output is None:
    #     assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
    #     raise NotImplementedError

    #     # out_path = top_level.parent / (top_level.stem + ".core_desc")
    # else:
    assert args.output is not None
    out_path = pathlib.Path(args.output)

    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    def process_metrics(settings, model=None):
        metrics = settings.metrics
        assert metrics is not None
        assert isinstance(metrics, list)

        def traverse(x, prefix=None):
            assert isinstance(x, dict)
            assert len(x) == 1
            name = list(x.keys())[0]
            prefix = name if prefix is None else f"{prefix}.{name}"
            y = list(x.values())[0]
            assert isinstance(y, dict)
            y = y.copy()
            passes = y.pop("passes", None)
            if passes is None:
                passes = []
            assert isinstance(passes, list)
            ret = {}
            for pass_ in passes:
                ret_ = traverse(pass_, prefix=prefix)
                ret.update(ret_)
            models = y.pop("models", None)
            if models is None:
                models = []
            assert isinstance(models, list)
            for model in models:
                ret2_ = traverse(model, prefix=prefix)
                ret.update(ret2_)
            for key, value in y.items():
                ret[f"{prefix}.{key}"] = value
            return ret

        all_metrics = {}
        for stage_metrics in metrics:
            result = traverse(stage_metrics)
            all_metrics.update(result)

        def filter_func(x):
            return x.split(".")[-1] in ["success_instructions", "failed_instructions", "skipped_instructions"]

        filtered_metrics = {key: val for key, val in all_metrics.items() if filter_func(key)}
        if model:
            filtered_metrics = {key.replace(f"{model}.", ""): val for key, val in filtered_metrics.items()}
        return filtered_metrics

    def get_status(filtered_metrics, instr_name, invert: bool = False):
        skipped = []
        failed = []
        success = []
        for key, val in filtered_metrics.items():
            if not isinstance(val, list):
                continue
            if instr_name not in val:
                continue
            prefix, metric = key.rsplit(".", 1)
            if metric == "skipped_instructions":
                skipped.append(prefix)
            elif metric == "failed_instructions":
                failed.append(prefix)
            elif metric == "success_instructions":
                success.append(prefix)
        if invert:
            ret = {}
            for pass_name in failed:
                ret[pass_name] = "failed"
            for pass_name in skipped:
                if pass_name in ret:
                    continue
                ret[pass_name] = "skipped"
            for pass_name in success:
                if pass_name in ret:
                    continue
                ret[pass_name] = "success"
        else:
            ret = {
                "skipped": skipped,
                "failed": failed,
                "success": success,
            }
        return ret

    status_data = []

    # resolve model paths
    top_levels = args.top_level
    for top_level in top_levels:
        top_level = pathlib.Path(top_level)
        # abs_top_level = top_level.resolve()

        model_obj = load_model(top_level, compat=args.compat)

        for set_name, set_def in model_obj.sets.items():
            if len(set_def.instructions) == 0:
                continue
            xlen = set_def.xlen
            model = top_level.stem
            filtered_metrics = process_metrics(settings, model=model)

            for instr_def in set_def.instructions.values():
                instr_name = instr_def.name

                data = {
                    "model": model,
                    "set": set_name,
                    "xlen": xlen,
                    "instr": instr_name,
                }
                if not args.compact:
                    for pass_name, pass_status in get_status(filtered_metrics, instr_name, invert=True).items():
                        stage_name, pass_rest = pass_name.split(".", 1)
                        data_ = {**data, "stage": stage_name, "pass": pass_rest, "status": pass_status}
                        status_data.append(data_)
                else:
                    data_ = {**data, **get_status(filtered_metrics, instr_name, invert=False)}
                    data_["n_success"] = len(data_["success"])
                    del data_["success"]
                    data_["n_skipped"] = len(data_["skipped"])
                    del data_["skipped"]
                    data_["n_failed"] = len(data_["failed"])
                    del data_["failed"]
                    data_["n_total"] = data_["n_failed"] + data_["n_skipped"] + data_["n_success"]

                    def helper(success, skipped, failed):
                        if failed:
                            return "bad"
                        if skipped:
                            return "ok"
                        if success:
                            return "good"
                        return "unknown"

                    data_["status"] = helper(data_["n_success"], data_["n_skipped"], data_["n_failed"])
                    status_data.append(data_)
    status_df = pd.DataFrame(status_data)
    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        status_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        status_df.to_pickle(out_path)
    elif fmt == "md":
        # status_df.to_markdown(out_path, tablefmt="grid", index=False)
        if args.markdown_icons:

            def helper2(x):
                x = x.replace("success", "success :heavy_check_mark:")
                x = x.replace("skipped", "skipped :question:")
                x = x.replace("failed", "failed :x:")
                x = x.replace("good", "good :green_circle:")
                x = x.replace("ok", "ok :orange_circle:")
                x = x.replace("unknown", "unknown :yellow_circle:")
                x = x.replace("bad", "bad :red_circle:")
                return x

            status_df["status"] = status_df["status"].apply(helper2)
        status_df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Very crude expression evaluation functions for use during model generation."""

from m2isar.metamodel import arch, behav


def group(self: behav.Group, context):
    return self.expr.generate(context)


def int_literal(self: behav.IntLiteral, context):
    return self.value


def named_reference(self: behav.NamedReference, context):
    if isinstance(self.reference, arch.Constant) and self.reference.value is not None:
        return self.reference.value
    return None
    # raise M2ValueError("non-interpretable value encountered")


def indexed_reference(self: behav.IndexedReference, context):
    idx = self.index.generate(context)
    return self.reference._initval[idx]


def binary_operation(self: behav.BinaryOperation, context):
    left = self.left.generate(context)
    right = self.right.generate(context)
    if left is None or right is None:
        return None
    return int(eval(f"{left}{self.op.value}{right}"))


def unary_operation(self: behav.UnaryOperation, context):
    right = self.right.generate(context)
    return int(eval(f"{self.op.value}{right}"))



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for resetting the seal5 environment."""

from seal5.flow import Seal5Flow
from seal5.logging import Logger


logger = Logger("cli")


def add_reset_options(parser):
    """Setup parser for reset argument group."""
    reset_parser = parser.add_argument_group("reset options")
    reset_parser.add_argument(
        "--non-interactive",
        default=True,
        action="store_true",
        help="Do not ask questions interactively",
    )
    reset_parser.add_argument(
        "--settings",
        default=False,
        dest="settings",
        action="store_true",
        help="Should settings be reset?",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the reset subcommand."""
    parser = subparsers.add_parser("reset", description="Reset Seal5 settings.")
    parser.set_defaults(func=handle)
    add_reset_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the reset subcommand"""
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.reset(settings=args.settings, verbose=args.verbose, interactive=not args.non_interactive)



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Utility stuff for M2-ISA-R viewer"""

import tkinter as tk
from tkinter import ttk
from anytree import Node


class TreeGenContext:
    def __init__(self, parent=None) -> None:
        if parent:
            self.nodes = [parent]
        else:
            self.nodes = [Node("Tree")]
        self.parent_stack = [0]
        # self.layer = 0

    @property
    def parent(self):
        return self.nodes[self.parent_stack[-1]]

    @property
    def tree(self):
        return self.nodes[0]

    def push(self, node):
        # self.layer += 1
        self.parent_stack.append(len(self.nodes))
        self.nodes.append(node)

    def pop(self):
        # self.layer -= 1
        return self.parent_stack.pop()

    def insert(self, text, values=None):
        self.push(self.insert2(text, values=values))

    def insert2(self, text, values=None):
        raise NotImplementedError


class TkTreeGenContext(TreeGenContext):
    def __init__(self, tree: "ttk.Treeview", parent=None) -> None:
        super().__init__(parent=parent)
        self.tree = tree

    def insert2(self, text, values=None):
        return self.tree.insert(self.parent, tk.END, text=text, values=values)


class TextTreeGenContext(TreeGenContext):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)

    def insert2(self, text, values=None):
        if isinstance(values, (list, set, tuple)):
            if len(values) == 1:
                values = values[0]
        return Node(text, parent=self.parent, value=values)



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Generate a ttk.Treeview representation of a M2-ISA-R model structure."""

# import tkinter as tk

from m2isar.metamodel import behav
from .utils import TreeGenContext

# pylint: disable=unused-argument


def operation(self: behav.Operation, context: "TreeGenContext"):
    context.insert("Operation")

    for stmt in self.statements:
        if stmt is None:
            continue
        stmt.generate(context)

    context.pop()


def block(self: behav.Block, context: "TreeGenContext"):
    context.insert("Block")

    for stmt in self.statements:
        if stmt is None:
            continue
        stmt.generate(context)

    context.pop()


def binary_operation(self: behav.BinaryOperation, context: "TreeGenContext"):
    context.insert("Binary Operation")

    context.insert("Left")
    self.left.generate(context)
    context.pop()

    context.insert("Right")
    self.right.generate(context)
    context.pop()

    context.insert2("Op", values=(self.op.value,))

    context.pop()


def slice_operation(self: behav.SliceOperation, context: "TreeGenContext"):
    context.insert("Slice Operation")

    context.insert("Expr")
    self.expr.generate(context)
    context.pop()

    context.insert("Left")
    self.left.generate(context)
    context.pop()

    context.insert("Right")
    self.right.generate(context)
    context.pop()

    context.pop()


def concat_operation(self: behav.ConcatOperation, context: "TreeGenContext"):
    context.insert("Concat Operation")

    context.insert("Left")
    self.left.generate(context)
    context.pop()

    context.insert("Right")
    self.right.generate(context)
    context.pop()

    context.pop()


def number_literal(self: behav.IntLiteral, context: "TreeGenContext"):
    context.insert2("Number Literal", values=(self.value,))


def int_literal(self: behav.IntLiteral, context: "TreeGenContext"):
    context.insert2("Int Literal", values=(self.value,))


def scalar_definition(self: behav.ScalarDefinition, context: "TreeGenContext"):
    context.insert2("Scalar Definition", values=(self.scalar.name,))


def break_(self: behav.Break, context: "TreeGenContext"):
    context.insert2("Break")


def assignment(self: behav.Assignment, context: "TreeGenContext"):
    context.insert("Assignment")

    context.insert("Target")
    self.target.generate(context)
    context.pop()

    context.insert("Expr")
    self.expr.generate(context)
    context.pop()

    context.pop()


def conditional(self: behav.Conditional, context: "TreeGenContext"):
    context.insert("Conditional")

    context.insert("Conditions")
    for cond in self.conds:
        cond.generate(context)
    context.pop()

    context.insert("Statements")
    for stmt in self.stmts:
        if stmt is None:
            continue
        stmt.generate(context)
    context.pop()

    context.pop()


def loop(self: behav.Loop, context: "TreeGenContext"):
    context.insert("Loop")

    context.insert2("Post Test", values=(self.post_test,))

    context.insert("Condition")
    self.cond.generate(context)
    context.pop()

    context.insert("Statements")
    for stmt in self.stmts:
        stmt.generate(context)
    context.pop()

    context.pop()


def ternary(self: behav.Ternary, context: "TreeGenContext"):
    context.insert("Ternary")

    context.insert("Cond")
    self.cond.generate(context)
    context.pop()

    context.insert("Then Expression")
    self.then_expr.generate(context)
    context.pop()

    context.insert("Else Expression")
    self.else_expr.generate(context)
    context.pop()

    context.pop()


def return_(self: behav.Return, context: "TreeGenContext"):
    context.insert("Return")

    if self.expr is not None:
        context.insert("Expression")
        self.expr.generate(context)
        context.pop()

    context.pop()


def unary_operation(self: behav.UnaryOperation, context: "TreeGenContext"):
    context.insert("Unary Operation")

    context.insert("Right")
    self.right.generate(context)
    context.pop()

    context.insert2("Op", values=(self.op.value,))

    context.pop()


def named_reference(self: behav.NamedReference, context: "TreeGenContext"):
    context.insert2("Named Reference", values=(f"{self.reference}",))


def indexed_reference(self: behav.IndexedReference, context: "TreeGenContext"):
    context.insert("Indexed Reference")

    context.insert2("Reference", values=(f"{self.reference}",))

    context.insert("Index")
    self.index.generate(context)
    context.pop()

    context.pop()


def type_conv(self: behav.TypeConv, context: "TreeGenContext"):
    context.insert("Type Conv")

    context.insert2("Type", values=(self.data_type,))
    context.insert2("Size", values=(self.size,))

    context.insert("Expr")
    self.expr.generate(context)
    context.pop()

    context.pop()


def callable_(self: behav.Callable, context: "TreeGenContext"):
    context.insert("Callable", values=(self.ref_or_name.name,))

    for arg, arg_descr in zip(self.args, self.ref_or_name.args):
        context.insert("Arg", values=(arg_descr,))
        arg.generate(context)
        context.pop()

    context.pop()


def group(self: behav.Group, context: "TreeGenContext"):
    context.insert("Group")

    context.insert("Expr")
    self.expr.generate(context)
    context.pop()

    context.pop()



# Generated from CoreDSL2.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .CoreDSL2Parser import CoreDSL2Parser
else:
    from CoreDSL2Parser import CoreDSL2Parser

# This class defines a complete generic visitor for a parse tree produced by CoreDSL2Parser.

class CoreDSL2Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by CoreDSL2Parser#description_content.
    def visitDescription_content(self, ctx:CoreDSL2Parser.Description_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#import_file.
    def visitImport_file(self, ctx:CoreDSL2Parser.Import_fileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#instruction_set.
    def visitInstruction_set(self, ctx:CoreDSL2Parser.Instruction_setContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#core_def.
    def visitCore_def(self, ctx:CoreDSL2Parser.Core_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#section_arch_state.
    def visitSection_arch_state(self, ctx:CoreDSL2Parser.Section_arch_stateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#section_functions.
    def visitSection_functions(self, ctx:CoreDSL2Parser.Section_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#section_instructions.
    def visitSection_instructions(self, ctx:CoreDSL2Parser.Section_instructionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#section_always.
    def visitSection_always(self, ctx:CoreDSL2Parser.Section_alwaysContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#always_block.
    def visitAlways_block(self, ctx:CoreDSL2Parser.Always_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#instruction.
    def visitInstruction(self, ctx:CoreDSL2Parser.InstructionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#rule_encoding.
    def visitRule_encoding(self, ctx:CoreDSL2Parser.Rule_encodingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#bit_value.
    def visitBit_value(self, ctx:CoreDSL2Parser.Bit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#bit_field.
    def visitBit_field(self, ctx:CoreDSL2Parser.Bit_fieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#function_definition.
    def visitFunction_definition(self, ctx:CoreDSL2Parser.Function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#parameter_list.
    def visitParameter_list(self, ctx:CoreDSL2Parser.Parameter_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#parameter_declaration.
    def visitParameter_declaration(self, ctx:CoreDSL2Parser.Parameter_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#block_statement.
    def visitBlock_statement(self, ctx:CoreDSL2Parser.Block_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#procedure_call.
    def visitProcedure_call(self, ctx:CoreDSL2Parser.Procedure_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#if_statement.
    def visitIf_statement(self, ctx:CoreDSL2Parser.If_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#for_statement.
    def visitFor_statement(self, ctx:CoreDSL2Parser.For_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#while_statement.
    def visitWhile_statement(self, ctx:CoreDSL2Parser.While_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#do_statement.
    def visitDo_statement(self, ctx:CoreDSL2Parser.Do_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#switch_statement.
    def visitSwitch_statement(self, ctx:CoreDSL2Parser.Switch_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#return_statement.
    def visitReturn_statement(self, ctx:CoreDSL2Parser.Return_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#break_statement.
    def visitBreak_statement(self, ctx:CoreDSL2Parser.Break_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#continue_statement.
    def visitContinue_statement(self, ctx:CoreDSL2Parser.Continue_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#spawn_statement.
    def visitSpawn_statement(self, ctx:CoreDSL2Parser.Spawn_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#expression_statement.
    def visitExpression_statement(self, ctx:CoreDSL2Parser.Expression_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#switch_block_statement_group.
    def visitSwitch_block_statement_group(self, ctx:CoreDSL2Parser.Switch_block_statement_groupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#switch_label.
    def visitSwitch_label(self, ctx:CoreDSL2Parser.Switch_labelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#block.
    def visitBlock(self, ctx:CoreDSL2Parser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#block_item.
    def visitBlock_item(self, ctx:CoreDSL2Parser.Block_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#for_condition.
    def visitFor_condition(self, ctx:CoreDSL2Parser.For_conditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#declaration.
    def visitDeclaration(self, ctx:CoreDSL2Parser.DeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#type_specifier.
    def visitType_specifier(self, ctx:CoreDSL2Parser.Type_specifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#integer_type.
    def visitInteger_type(self, ctx:CoreDSL2Parser.Integer_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#float_type.
    def visitFloat_type(self, ctx:CoreDSL2Parser.Float_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#bool_type.
    def visitBool_type(self, ctx:CoreDSL2Parser.Bool_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#void_type.
    def visitVoid_type(self, ctx:CoreDSL2Parser.Void_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#composite_declaration.
    def visitComposite_declaration(self, ctx:CoreDSL2Parser.Composite_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#composite_reference.
    def visitComposite_reference(self, ctx:CoreDSL2Parser.Composite_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#enum_declaration.
    def visitEnum_declaration(self, ctx:CoreDSL2Parser.Enum_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#enum_reference.
    def visitEnum_reference(self, ctx:CoreDSL2Parser.Enum_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#integer_signedness.
    def visitInteger_signedness(self, ctx:CoreDSL2Parser.Integer_signednessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#integer_shorthand.
    def visitInteger_shorthand(self, ctx:CoreDSL2Parser.Integer_shorthandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#float_shorthand.
    def visitFloat_shorthand(self, ctx:CoreDSL2Parser.Float_shorthandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#attribute.
    def visitAttribute(self, ctx:CoreDSL2Parser.AttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#bit_size_specifier.
    def visitBit_size_specifier(self, ctx:CoreDSL2Parser.Bit_size_specifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#enumerator_list.
    def visitEnumerator_list(self, ctx:CoreDSL2Parser.Enumerator_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#enumerator.
    def visitEnumerator(self, ctx:CoreDSL2Parser.EnumeratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#struct_declaration.
    def visitStruct_declaration(self, ctx:CoreDSL2Parser.Struct_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#struct_declaration_specifier.
    def visitStruct_declaration_specifier(self, ctx:CoreDSL2Parser.Struct_declaration_specifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#declarator.
    def visitDeclarator(self, ctx:CoreDSL2Parser.DeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#initializer.
    def visitInitializer(self, ctx:CoreDSL2Parser.InitializerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#initializerList.
    def visitInitializerList(self, ctx:CoreDSL2Parser.InitializerListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#designated_initializer.
    def visitDesignated_initializer(self, ctx:CoreDSL2Parser.Designated_initializerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#designator.
    def visitDesignator(self, ctx:CoreDSL2Parser.DesignatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#cast_expression.
    def visitCast_expression(self, ctx:CoreDSL2Parser.Cast_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#binary_expression.
    def visitBinary_expression(self, ctx:CoreDSL2Parser.Binary_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#preinc_expression.
    def visitPreinc_expression(self, ctx:CoreDSL2Parser.Preinc_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#conditional_expression.
    def visitConditional_expression(self, ctx:CoreDSL2Parser.Conditional_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#deref_expression.
    def visitDeref_expression(self, ctx:CoreDSL2Parser.Deref_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#prefix_expression.
    def visitPrefix_expression(self, ctx:CoreDSL2Parser.Prefix_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#postinc_expression.
    def visitPostinc_expression(self, ctx:CoreDSL2Parser.Postinc_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#concat_expression.
    def visitConcat_expression(self, ctx:CoreDSL2Parser.Concat_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#assignment_expression.
    def visitAssignment_expression(self, ctx:CoreDSL2Parser.Assignment_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#method_call.
    def visitMethod_call(self, ctx:CoreDSL2Parser.Method_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#primary_expression.
    def visitPrimary_expression(self, ctx:CoreDSL2Parser.Primary_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#slice_expression.
    def visitSlice_expression(self, ctx:CoreDSL2Parser.Slice_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#reference_expression.
    def visitReference_expression(self, ctx:CoreDSL2Parser.Reference_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#constant_expression.
    def visitConstant_expression(self, ctx:CoreDSL2Parser.Constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#literal_expression.
    def visitLiteral_expression(self, ctx:CoreDSL2Parser.Literal_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#parens_expression.
    def visitParens_expression(self, ctx:CoreDSL2Parser.Parens_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#string_literal.
    def visitString_literal(self, ctx:CoreDSL2Parser.String_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#constant.
    def visitConstant(self, ctx:CoreDSL2Parser.ConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#integer_constant.
    def visitInteger_constant(self, ctx:CoreDSL2Parser.Integer_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#floating_constant.
    def visitFloating_constant(self, ctx:CoreDSL2Parser.Floating_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#bool_constant.
    def visitBool_constant(self, ctx:CoreDSL2Parser.Bool_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#character_constant.
    def visitCharacter_constant(self, ctx:CoreDSL2Parser.Character_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#string_constant.
    def visitString_constant(self, ctx:CoreDSL2Parser.String_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#double_left_bracket.
    def visitDouble_left_bracket(self, ctx:CoreDSL2Parser.Double_left_bracketContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#double_right_bracket.
    def visitDouble_right_bracket(self, ctx:CoreDSL2Parser.Double_right_bracketContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#data_types.
    def visitData_types(self, ctx:CoreDSL2Parser.Data_typesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#type_qualifier.
    def visitType_qualifier(self, ctx:CoreDSL2Parser.Type_qualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#storage_class_specifier.
    def visitStorage_class_specifier(self, ctx:CoreDSL2Parser.Storage_class_specifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CoreDSL2Parser#struct_or_union.
    def visitStruct_or_union(self, ctx:CoreDSL2Parser.Struct_or_unionContext):
        return self.visitChildren(ctx)



del CoreDSL2Parser


# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""
from functools import reduce

from enum import IntFlag, auto

from m2isar.metamodel import arch, behav

# pylint: disable=unused-argument

# Limitations
# - Keep the condition stack as small as possible (move raise to highest level of behavior)
# - If block should only have a call to raise
# - Standalone if statements (without else/else if) are preferred
# - Raise calls are replaced by empty blocks, to be removed in the operation visitor
# - ~~Run optimizer afterwards to eliminate if statement.~~


class Mode(IntFlag):
    NONE = 0
    REG = auto()
    IMM = auto()
    CONST = auto()
    IMM_CONST = IMM | CONST


def operation(self: behav.Operation, context):
    # print("operation", self)
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            for _t in temp:
                pass
                # print("t", t, type(t), dir(t))
            statements.extend(temp)
        else:
            # print("temp", temp, type(temp), dir(temp))
            statements.append(temp)
    # input("eeeee")

    return reduce(lambda x, y: x | y, statements)


def block(self: behav.Block, context):
    # print("block")

    stmts = []

    for stmt in self.statements:
        stmt = stmt.generate(context)
        if isinstance(stmt, behav.Conditional):
            if len(stmt.conds) == 1:
                if isinstance(stmt.stmts[0], behav.Block):
                    if len(stmt.stmts[0].statements) == 0:
                        continue
        stmts.append(stmt)

    return reduce(lambda x, y: x | y, stmts) if len(stmts) > 0 else Mode.NONE


def binary_operation(self: behav.BinaryOperation, context):
    # print("binary_operation {")
    left = self.left.generate(context)
    right = self.right.generate(context)
    # print("left", left)
    # print("right", right)
    # print("left | right", left | right)
    # print("(left | right) == IMM_CONST", (left | right) == Mode.IMM_CONST)
    if (left | right) == Mode.IMM_CONST:
        assert context.last_imm_name is not None
        context.imm_leaf_names.add(context.last_imm_name)
        context.last_imm_name = None
        return Mode.NONE
    # input("} binary_operation")

    return left | right


def slice_operation(self: behav.SliceOperation, context):
    # print("slice_operation")
    expr = self.expr.generate(context)
    left = self.left.generate(context)
    right = self.right.generate(context)

    return expr | left | right


def concat_operation(self: behav.ConcatOperation, context):
    # print("concat_operation")
    left = self.left.generate(context)
    right = self.right.generate(context)

    return left | right


def number_literal(self: behav.IntLiteral, context):
    # print("number_literal")
    return Mode.CONST


def int_literal(self: behav.IntLiteral, context):
    # print("int_literal")
    return Mode.CONST


def scalar_definition(self: behav.ScalarDefinition, context):
    # print("scalar_definition")
    return Mode.NONE


def break_(self: behav.Break, context):
    # print("break")
    return Mode.NONE


def assignment(self: behav.Assignment, context):
    # print("assignment")
    target = self.target.generate(context)
    expr = self.expr.generate(context)

    return target | expr


def conditional(self: behav.Conditional, context):
    # print("conditional")
    conds = []
    for i, cond in enumerate(self.conds):
        cond_ = cond.generate(context)
        conds.append(cond_)
    stmts = []
    for op in self.stmts:
        # for stmt in flatten(op):
        #    stmt.generate(context)
        stmt = op.generate(context)
        stmts.append(stmt)

    return reduce(lambda x, y: x | y, conds + stmts)


def loop(self: behav.Loop, context):
    # print("loop")
    cond = self.cond.generate(context)
    stmts = [x.generate(context) for x in self.stmts]

    return reduce(lambda x, y: x | y, [cond] + stmts)


def ternary(self: behav.Ternary, context):
    # print("ternary")
    cond = self.cond.generate(context)
    then_expr = self.then_expr.generate(context)
    else_expr = self.else_expr.generate(context)

    return cond | then_expr | else_expr


def return_(self: behav.Return, context):
    # print("return_")
    expr = Mode.NONE
    if self.expr is not None:
        expr = self.expr.generate(context)

    return expr


def unary_operation(self: behav.UnaryOperation, context):
    # print("unary_operation {")
    right = self.right.generate(context)
    # print("right", right)
    # print("right == IMM_CONST", right == Mode.IMM_CONST)
    # input("} unary_operation")
    if right == Mode.IMM_CONST:
        assert context.last_imm_name is not None
        context.imm_leaf_names.add(context.last_imm_name)
        context.last_imm_name = None
        return Mode.NONE

    return right


def named_reference(self: behav.NamedReference, context):
    # print("named_reference")
    # print("name", self.reference.name)
    # TODO: hanlde constant and scalar!
    if isinstance(self.reference, arch.BitFieldDescr):
        # print("if")
        if self.reference.name in context.imm_op_names:
            # print("if if")
            # input(">>>")
            context.last_imm_name = self.reference.name
            return Mode.IMM
        else:
            return Mode.REG
    # print("return None")
    return Mode.NONE


def indexed_reference(self: behav.IndexedReference, context):
    # print("indexed_reference")
    assert isinstance(self.reference, arch.Memory)
    # name = self.reference.name

    index = self.index.generate(context)
    return index


def type_conv(self: behav.TypeConv, context):
    # print("type_conv {")
    expr = self.expr.generate(context)
    # print("expr", expr)
    # input("} type_conv")

    return expr


def callable_(self: behav.Callable, context):
    # print("callable")
    args = [stmt.generate(context) for stmt in self.args]

    return reduce(lambda x, y: x | y, args)


def group(self: behav.Group, context):
    # print("group {")
    expr = self.expr.generate(context)
    # print("expr", expr)
    # input("} group")

    return expr


def procedure_call(self: behav.ProcedureCall, context):
    # print("procedure_call")

    fn_args = [arg.generate(context) for arg in self.args]

    return reduce(lambda x, y: x | y, fn_args)



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import copy
import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import arch, behav, patch_model

from seal5.model_utils import load_model

from . import visitor

from seal5.logging import Logger

logger = Logger("backends.coredsl2_writer")


class CoreDSL2Writer:
    # def __init__(self, reduced=False):
    def __init__(self, reduced=True, version="coredsl2"):
        self.version = version
        self.reduced = reduced  # Reduced syntax for cdsl2llvm parser
        self.text = ""
        self.indent_str = "    "
        self.level = 0

    @property
    def is_seal5(self):
        # TODO: use this
        return self.version == "seal5"

    @property
    def is_coredsl2(self):
        return self.version == "coredsl2"

    @property
    def indent(self):
        return self.indent_str * self.level

    @property
    def isstartofline(self):
        return len(self.text) == 0 or self.text[-1] == "\n"

    @property
    def needsspace(self):
        return len(self.text) != 0 and self.text[-1] not in ["\n", " "]

    def write(self, text, nl=False):
        # print("text", text, type(text))
        if isinstance(text, int):
            text = str(text)
        assert isinstance(text, str)
        lines = text.split("\n")
        for i, line in enumerate(lines):
            if self.isstartofline:
                self.text += self.indent
            self.text += line
            if (i < len(lines) - 1) or nl:
                self.text += "\n"

    def write_line(self, text):
        self.write(text, nl=True)

    def enter_block(self, br=True, nl=True):
        if br:
            if self.needsspace:
                self.write(" ")
            self.write("{", nl=nl)
        self.level += 1

    def leave_block(self, br=True, nl=True):
        assert self.level > 0
        self.level -= 1
        if br:
            self.write("}", nl=nl)

    def write_type(self, data_type, size):
        # print("write_type")
        # print("data_type", data_type)
        # print("size", size)
        if data_type == arch.DataType.U:
            self.write("unsigned")
        elif data_type == arch.DataType.S:
            self.write("signed")
        elif data_type == arch.DataType.NONE:
            self.write("void")
        else:
            raise NotImplementedError(f"Unsupported type: {data_type}")
        if size:
            self.write("<")
            self.write(size)
            self.write(">")

    def write_attribute(self, attr, val=None):
        if self.needsspace:
            self.write(" ")
        if val is not None:
            if isinstance(val, list) and len(val) == 0:
                val = None
        if self.reduced and val is not None:
            return
        # TODO: allow atrbitrary attrs in cdsl2llvm parser, not only for operands
        allowed_attrs = ["is_unsigned", "is_signed", "is_imm", "is_reg", "in", "out", "inout", "is_32_bit"]
        if self.reduced and attr.name.lower() not in allowed_attrs:
            return
        self.write("[[")
        self.write(attr.name.lower())
        if val is not None:
            self.write("=")

            def helper(val):
                if isinstance(val, list):  # TODO: replace with string literal
                    if len(val) == 1:
                        return helper(val[0])
                    return "(" + ",".join([helper(x) for x in val]) + ")"
                if isinstance(val, str):  # TODO: replace with string literal
                    if '"' not in val:
                        val = '"' + val + '"'
                    return val
                if isinstance(val, int):  # TODO: replace with int literal
                    return str(val)  # TODO: operation
                if isinstance(val, behav.IntLiteral):
                    size = val.bit_size
                    value = val.value
                    if size and value != 0 and not val.signed:
                        # hex mode
                        return f"{size}'h{value:x}"
                    else:
                        return str(val.value)
                if isinstance(val, behav.StringLiteral):
                    val = val.value
                    if '"' not in val:
                        val = '"' + val + '"'
                    return val
                raise NotImplementedError(f"Unhandled case: {type(val)}")

            val = helper(val)
            self.write(val)
        self.write("]]")
        # print("key", key)
        # print("value", value)

    def write_attributes(self, attributes):
        for attr, val in attributes.items():
            self.write_attribute(attr, val)
        # input("inp")

    def write_function(self, function):
        if function.static:
            self.write("static ")
        if function.extern:
            self.write("extern ")
        self.write_type(function.data_type, None)
        if function.size is not None:
            self.write("<")
            self.write(f"{function.size}")
            self.write(">")
        self.write(" ")
        self.write(function.name)
        self.write("(")
        for i, param in enumerate(function.args.values()):
            self.write_type(param.data_type, param.size)
            if param.name is not None:
                self.write(" ")
                self.write(param.name)
            if i < len(function.args) - 1:
                self.write(", ")
        self.write(")")
        self.write_attributes(function.attributes)
        # self.enter_block()
        # self.write_behavior(instruction)
        if function.extern:
            self.write_line(";")
        else:
            function.operation.generate(self)
        # self.leave_block()

    def write_functions(self, functions):
        if self.reduced:
            return
        self.write("functions")
        # TODO: attributes
        self.enter_block()
        for function in functions.values():
            self.write_function(function)
        self.leave_block()

    def write_encoding_val(self, bitval):
        value = bitval.value
        width = bitval.length
        self.write(width)
        self.write("'b")
        bitstr = bin(value)[2:].zfill(width)
        self.write(bitstr)

    def write_encoding_field(self, bitfield):
        name = bitfield.name
        rng = bitfield.range
        # print("rng", rng)
        # input("aaaa")
        self.write(name)
        self.write(f"[{rng.upper}:{rng.lower}]")

    def write_operand(self, operand):
        self.write_type(operand.ty.datatype, operand.ty.width)
        self.write(" ")
        self.write(operand.name)
        self.write_attributes(operand.attributes)
        self.write_line(";")

    def write_constraints(self, constraints):
        for constraint in constraints:
            # print("constraint", constraint, type(constraint), dir(constraint))
            for stmt in constraint.stmts:
                stmt.generate(self)
                desc = constraint.description
                if desc:
                    self.write(f";  // {desc}", nl=True)
                else:
                    self.write(";", nl=True)

    def write_instruction_constraints(self, constraints, operands):
        if self.reduced:
            return
        self.write("constraints: ")
        if len(constraints) == 0:
            self.write_line("{};")
            return
        self.enter_block()
        op_constraints = sum([op.constraints for op in operands.values()], [])
        self.write_constraints(op_constraints)
        self.write_constraints(constraints)
        self.leave_block()

    def write_operands(self, operands):
        self.write("operands: ")
        if len(operands) == 0:
            self.write_line("{}")
            return
        self.enter_block()
        # print("operands", operands)
        for _, op in enumerate(operands.values()):
            self.write_operand(op)
        # for i, op in enumerate(operands.values()):
        #     self.write_constraints(op.constraints)
        self.leave_block()

    def write_encoding(self, encoding):
        self.write("encoding: ")
        # print("encoding", encoding, dir(encoding))
        for i, elem in enumerate(encoding):
            if isinstance(elem, arch.BitVal):
                self.write_encoding_val(elem)
            elif isinstance(elem, arch.BitField):
                self.write_encoding_field(elem)
            else:
                assert False
            if i < len(encoding) - 1:
                self.write(" :: ")
        self.write(";", nl=True)

    def write_assembly(self, instruction):
        self.write("assembly: ")
        mnemonic = instruction.mnemonic
        assembly = instruction.assembly
        if mnemonic and not self.reduced:
            self.write("{")
            self.write(f'"{mnemonic}"')
            self.write(", ")
        if assembly is None:
            assembly = ""
        self.write(f'"{assembly}"')
        if mnemonic and not self.reduced:
            self.write("}")
        self.write(";", nl=True)

    def write_behavior(self, instruction):
        self.write("behavior: ")
        op = instruction.operation
        if self.reduced:
            self.enter_block()
        op.generate(self)
        if self.reduced:
            self.leave_block()
        # self.write(";", nl=True)

    def write_instruction(self, instruction):
        # print("write_instruction", instruction)
        self.write(instruction.name)
        self.write_attributes(instruction.attributes)
        self.enter_block()
        self.write_operands(instruction.operands)  # seal5 only
        self.write_instruction_constraints(instruction.constraints, instruction.operands)  # seal5 only
        self.write_encoding(instruction.encoding)
        self.write_assembly(instruction)
        self.write_behavior(instruction)
        self.leave_block()

    def write_instructions(self, instructions):
        # print("write_instructions", instructions)
        self.write("instructions")
        # TODO: attributes?
        self.enter_block()
        for instruction in instructions.values():
            self.write_instruction(instruction)
        self.leave_block()

    def write_architectural_state(self, _set_def):
        self.write("architectural_state")
        # TODO: use set_def
        # print("set_def", set_def, dir(set_def))
        self.enter_block()
        # TODO: scalars, memories,...
        self.leave_block()

    def write_set(self, set_def):
        # print("write_set", set_def)
        # self.write_architectural_state()
        self.write("InstructionSet ")
        self.write(set_def.name)
        # TODO: attributes
        # TODO: extends
        if set_def.extension:
            self.write(" extends ")
            self.write(", ".join(set_def.extension))
        self.enter_block()
        self.write_functions(set_def.functions)
        self.write_instructions(set_def.instructions)
        self.leave_block()

    #     for instr_name, instr_def in set_def.instructions.items():
    #         logger.debug("writing instr %s", instr_def.name)
    #         # instr_def.operation.generate(context)
    #     # input("CONT1")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, required=True, default=None)
    parser.add_argument("--reduced", action="store_true", help="Generate pattern-gen compatible syntax")
    parser.add_argument("--splitted", action="store_true", help="Split per set and instruction")
    parser.add_argument("--ext", type=str, default="core_desc", help="Default file extension (if using --splitted)")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--ignore-failing", action="store_true", help="Do not crash in case of errors.")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    assert args.output is not None
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    # print("model", model)
    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
        "skipped_sets": [],
        "failed_sets": [],
        "success_sets": [],
    }
    if args.splitted:
        assert out_path.is_dir(), "Expecting output directory when using --splitted"
        for set_name, set_def in model_obj.sets.items():
            metrics["n_sets"] += 1
            for instr_def in set_def.instructions.values():
                metrics["n_instructions"] += 1
                writer = CoreDSL2Writer(reduced=args.reduced)
                logger.debug("writing instr %s/%s", set_def.name, instr_def.name)
                patch_model(visitor)
                set_def_ = copy.deepcopy(set_def)
                set_def_.instructions = {
                    key: instr_def
                    for key, instr_def_ in set_def.instructions.items()
                    if instr_def.name == instr_def_.name
                }
                try:
                    # TODO: drop_ununsed
                    writer.write_set(set_def_)
                    content = writer.text
                    out_path_ = out_path / set_name / f"{instr_def.name}.{args.ext}"
                    out_path_.parent.mkdir(exist_ok=True)
                    with open(out_path_, "w", encoding="utf-8") as f:
                        f.write(content)
                    metrics["n_success"] += 1
                    metrics["success_instructions"].append(instr_def.name)
                except Exception as ex:
                    logger.exception(ex)
                    metrics["n_failed"] += 1
                    metrics["failed_instructions"].append(instr_def.name)
    else:
        writer = CoreDSL2Writer(reduced=args.reduced)
        for set_name, set_def in model_obj.sets.items():
            metrics["n_sets"] += 1
            # print("set", set_def)
            # print("instrs", set_def.instructions)
            # input("123")
            logger.debug("writing set %s", set_def.name)
            patch_model(visitor)
            try:
                writer.write_set(set_def)
                metrics["n_success"] += 1
                metrics["success_sets"].append(set_name)
                # TODO: add instrs as well?
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_sets"].append(set_name)
        content = writer.text
        with open(out_path, "w", encoding="utf-8") as f:
            f.write(content)
    if not args.ignore_failing:
        n_failed = metrics["n_failed"]
        if n_failed > 0:
            failed = metrics["failed_instructions"]
            failing_str = ", ".join(failed)
            logger.error("%s intructions failed: %s", n_failed, failing_str)
            raise RuntimeError("Abort due to errors")
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

import antlr4
import antlr4.error.ErrorListener

from m2isar import M2SyntaxError
from .parser_gen import CoreDSL2Lexer, CoreDSL2Parser

RADIX = {"b": 2, "h": 16, "d": 10, "o": 8}

SHORTHANDS = {"char": 8, "short": 16, "int": 32, "long": 64}

SIGNEDNESS = {"signed": True, "unsigned": False}

BOOLCONST = {"true": 1, "false": 0}


class MyErrorListener(antlr4.error.ErrorListener.ErrorListener):
    def __init__(self, filename=None) -> None:
        self.filename = filename
        super().__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise M2SyntaxError(f"Syntax error in file {self.filename}, line {line}, column {column}: {msg}")


def make_parser(filename):
    input_stream = antlr4.FileStream(filename)
    lexer = CoreDSL2Lexer(input_stream)
    stream = antlr4.CommonTokenStream(lexer)
    parser = CoreDSL2Parser(stream)
    error_handler = MyErrorListener(filename)
    parser.removeErrorListeners()
    parser.addErrorListener(error_handler)
    return parser



"""Utilities for loading and dumping metamodels."""

from pathlib import Path
import pickle
from typing import Union

from m2isar.metamodel import M2_METAMODEL_VERSION, M2Model
from seal5.model import Seal5Model, SEAL5_METAMODEL_VERSION

from seal5.logging import Logger

logger = Logger("model_utils")


def load_model(
    model_path: Union[str, Path], compat: bool = False, allow_missmatch: bool = False
) -> Union[Seal5Model, M2Model]:
    logger.debug("loading model: %s", str(model_path))
    with open(model_path, "rb") as f:
        # models: "dict[str, arch.CoreDef]" = pickle.load(f)
        # sets: "dict[str, arch.InstructionSet]" = pickle.load(f)
        model_obj: Union[Seal5Model, M2Model] = pickle.load(f)
    if compat:
        assert isinstance(model_obj, M2Model), "Expected M2Model"
    else:
        assert isinstance(model_obj, Seal5Model), "Expected Seal5Model"
    required_version = M2_METAMODEL_VERSION if compat else SEAL5_METAMODEL_VERSION
    if model_obj.model_version != required_version:
        err_handler = logger.warning if allow_missmatch else RuntimeError
        err_handler("Loaded model version mismatch")
    return model_obj


def dump_model(
    model_obj: Union[Seal5Model, M2Model], out_path: Union[str, Path], compat: bool = False, ignore_suffix: bool = False
):
    if not ignore_suffix:
        out_path = Path(out_path)
        suffix = out_path.suffix
        required_suffix = ".m2isarmodel" if compat else ".seal5model"
        if suffix not in [".m2isarmodel", ".seal5model"]:
            assert len(suffix) == 0, f"Invalid suffix: {suffix}"
            out_path = out_path.parent / f"{out_path.stem}{required_suffix}"
        else:
            assert suffix == required_suffix, f"Invalid suffix: {suffix}, Expected: {required_suffix}"
    logger.debug("dumping model: %s", out_path)
    with open(out_path, "wb") as f:
        pickle.dump(model_obj, f)



"""Wrappers for running the seal5 flow with minimal efforts."""

import os
from typing import Optional, Union, List

# import logging
from pathlib import Path

from seal5.logging import Logger

from seal5.flow import Seal5Flow
from seal5.types import PatchStage
from seal5.utils import str2bool
from seal5.logging import update_log_level


logger = Logger("wrapper")


def group_files(files: List[Union[str, Path]]):
    """Split the given files per extension."""
    files = list(map(Path, files))
    cdsl_files = []
    cfg_files = []
    test_files = []
    other_files = []
    for file in files:
        suffix = file.suffix.lower()
        if suffix in [".core_desc", ".cdsl"]:
            cdsl_files.append(file)
        elif suffix in [".yml", ".yaml"]:
            cfg_files.append(file)
        elif suffix in [".c", ".cc", ".cpp", ".ll", ".mir", ".s"]:
            test_files.append(file)
        else:
            other_files.append(file)
    return cdsl_files, cfg_files, test_files, other_files


def prepatched_helper(val):
    if isinstance(val, str) and val.lower() == "auto":
        return "auto"
    return str2bool(val)


# Parameters
OUT_DIR = os.environ.get("OUT_DIR", None)
INSTALL_DIR = os.environ.get("INSTALL_DIR", None)
VERBOSE = str2bool(os.environ.get("VERBOSE", 0))
SKIP_PATTERNS = str2bool(os.environ.get("SKIP_PATTERNS", 0))
INTERACTIVE = str2bool(os.environ.get("INTERACTIVE", 0))
PREPATCHED = prepatched_helper(os.environ.get("PREPATCHED", "auto"))  # Possible values: [0,1,auto]
LLVM_URL = os.environ.get("LLVM_URL", "https://github.com/llvm/llvm-project.git")
LLVM_REF = os.environ.get("LLVM_REF", "llvmorg-19.1.7")
BUILD_CONFIG = os.environ.get("BUILD_CONFIG", None)
IGNORE_ERROR = str2bool(os.environ.get("IGNORE_ERROR", 1))
IGNORE_LLVM_IMM_TYPES = str2bool(os.environ.get("IGNORE_LLVM_IMM_TYPES", 0))
LOAD = str2bool(os.environ.get("LOAD", 1))
TRANSFORM = str2bool(os.environ.get("TRANSFORM", 1))
GENERATE = str2bool(os.environ.get("GENERATE", 1))
PATCH = str2bool(os.environ.get("PATCH", 1))
BUILD = str2bool(os.environ.get("BUILD", 1))
TEST = str2bool(os.environ.get("TEST", 1))
INSTALL = str2bool(os.environ.get("INSTALL", 1))
DEPLOY = str2bool(os.environ.get("DEPLOY", 1))
EXPORT = str2bool(os.environ.get("EXPORT", 1))
CLEANUP = str2bool(os.environ.get("CLEANUP", 0))
RESET = str2bool(os.environ.get("RESET", 1))
INIT = str2bool(os.environ.get("INIT", 1))
SETUP = str2bool(os.environ.get("SETUP", 1))
PROGRESS = str2bool(os.environ.get("PROGRESS", 1))
CCACHE = str2bool(os.environ.get("CCACHE", 0))
BUILD_CACHE = str2bool(os.environ.get("BUILD_CACHE", 0))
CLONE_DEPTH = int(os.environ.get("CLONE_DEPTH", -1))
LOG_LEVEL = os.environ.get("SEAL5_LOG_LEVEL", None)
NAME = os.environ.get("NAME", None)


def run_seal5_flow(
    input_files: List[Union[str, Path]],
    dest: Optional[Union[str, Path]] = None,
    out_dir: Optional[Union[str, Path]] = OUT_DIR,
    install_dir: Optional[Union[str, Path]] = INSTALL_DIR,
    name: Optional[str] = NAME,
    interactive: bool = INTERACTIVE,
    prepatched: bool = PREPATCHED,
    llvm_url: str = LLVM_URL,
    llvm_ref: str = LLVM_REF,
    verbose: bool = VERBOSE,
    clone_depth: Optional[int] = CLONE_DEPTH,
    progress: bool = PROGRESS,
    build_config: Optional[str] = BUILD_CONFIG,
    ccache: bool = CCACHE,
    enable_build_cache: bool = BUILD_CACHE,
    ignore_error: bool = IGNORE_ERROR,
    skip_patterns: bool = SKIP_PATTERNS,
    load: bool = LOAD,
    transform: bool = TRANSFORM,
    generate: bool = GENERATE,
    patch: bool = PATCH,
    build: bool = BUILD,
    test: bool = TEST,
    install: bool = INSTALL,
    deploy: bool = DEPLOY,
    export: bool = EXPORT,
    cleanup: bool = CLEANUP,
    reset: bool = RESET,
    init: bool = INIT,
    setup: bool = SETUP,
    ignore_llvm_imm_types: bool = IGNORE_LLVM_IMM_TYPES,
    log_level: Optional[str] = LOG_LEVEL,
):
    """Single entry point (wrapper) to excute the full seal5 flow for a given set of files."""
    seal5_flow = Seal5Flow(dest, name=name)

    # Optional: clean existing settings/models for fresh run
    if reset:
        seal5_flow.reset(settings=True, interactive=interactive)
        seal5_flow.clean(temp=True, patches=True, models=True, inputs=True, interactive=interactive)

    has_stage0_tag = not (seal5_flow.repo is None or f"seal5-{seal5_flow.name}-stage0" not in seal5_flow.repo.tags)
    if prepatched == "auto":
        prepatched = has_stage0_tag
    if prepatched:
        assert has_stage0_tag, "PREPATCHED can only be used after LLVM was patched at least once."
        logger.info("Skipping PHASE0 patch using PREPATCHED feature.")

    # Clone LLVM and init seal5 metadata directory
    if init:
        seal5_flow.initialize(
            clone=True,
            clone_url=llvm_url,
            clone_ref=f"seal5-{seal5_flow.name}-stage0" if prepatched else llvm_ref,
            clone_depth=clone_depth,
            progress=progress,
            force=True,
            verbose=verbose,
            ignore_llvm_imm_types=ignore_llvm_imm_types,
        )

    # Override log_level
    if log_level is not None:
        # seal5_flow.logger.parent.handlers[0].setLevel(log_level.upper())
        update_log_level(log_level)

    if load:
        if len(input_files) == 0:
            logger.warning("No input files provided")
        cdsl_files, cfg_files, test_files, other_files = group_files(input_files)

        # Load CoreDSL inputs
        seal5_flow.load(cdsl_files, verbose=verbose, overwrite=True)

        # Load test inputs
        seal5_flow.load(test_files, verbose=verbose, overwrite=True)

        # Load YAML inputs
        seal5_flow.load(cfg_files, verbose=verbose, overwrite=False)

        # Load other inputs
        seal5_flow.load(other_files, verbose=verbose, overwrite=False)

    # Override settings from Python
    if build_config is not None:
        seal5_flow.settings.llvm.default_config = build_config

    # Clone & install Seal5 dependencies
    # 1. CDSL2LLVM (add PHASE_0 patches)
    if setup:
        seal5_flow.setup(force=True, progress=progress, verbose=verbose, skip_patterns=skip_patterns)

    # Apply initial patches
    if not prepatched:
        seal5_flow.patch(verbose=verbose, stages=[PatchStage.PHASE_0])

    if build:
        # Build initial LLVM
        seal5_flow.build(
            verbose=verbose, config=build_config, enable_ccache=ccache, enable_build_cache=enable_build_cache
        )

    if transform:
        # Transform inputs
        #   1. Create M2-ISA-R metamodel
        #   2. Convert to Seal5 metamodel (including aliases, builtins,...)
        #   3. Analyse/optimize instructions
        seal5_flow.transform(verbose=verbose)

    if generate:
        # Generate patches (except Patterns)
        seal5_flow.generate(verbose=verbose, skip=["pattern_gen"])

    if patch:
        # Apply next patches
        seal5_flow.patch(verbose=verbose, stages=[PatchStage.PHASE_1, PatchStage.PHASE_2])

    if build:
        # Build patched LLVM
        seal5_flow.build(
            verbose=verbose, config=build_config, enable_ccache=ccache, enable_build_cache=enable_build_cache
        )

    if not skip_patterns:
        if build:
            # Build PatternGen & llc
            seal5_flow.build(
                verbose=verbose,
                config=build_config,
                target="pattern-gen",
                enable_ccache=ccache,
                enable_build_cache=enable_build_cache,
            )
            seal5_flow.build(
                verbose=verbose,
                config=build_config,
                target="llc",
                enable_ccache=ccache,
                enable_build_cache=enable_build_cache,
            )

        if generate:
            # Generate remaining patches
            seal5_flow.generate(verbose=VERBOSE, only=["pattern_gen"])

        if patch:
            # Apply patches
            seal5_flow.patch(verbose=verbose, stages=list(range(PatchStage.PHASE_3, PatchStage.PHASE_5 + 1)))

    if build:
        # Build patched LLVM
        seal5_flow.build(
            verbose=verbose, config=build_config, enable_ccache=ccache, enable_build_cache=enable_build_cache
        )

    if test:
        # Test patched LLVM
        seal5_flow.test(verbose=verbose, ignore_error=ignore_error)

    if install:
        # Install final LLVM
        if install_dir is None and out_dir is not None:
            install_dir = Path(out_dir) / "seal5_llvm_install"
        seal5_flow.install(dest=install_dir, verbose=verbose, config=build_config, enable_ccache=ccache)

    if deploy:
        # Deploy patched LLVM (export sources)
        # TODO: combine commits and create tag
        seal5_flow.deploy(
            f"{dest}_source.zip" if out_dir is None else Path(out_dir) / "seal5_llvm_source.zip", verbose=verbose
        )

    if export:
        # Export patches, logs, reports
        seal5_flow.export(f"{dest}.tar.gz" if out_dir is None else Path(out_dir) / "seal5.tar.gz", verbose=verbose)

    if cleanup:
        # Optional: cleanup temorary files, build dirs,...
        seal5_flow.clean(temp=True, patches=True, models=True, inputs=True, interactive=interactive)
